import logging
import sys
from json.decoder import JSONDecodeError
from typing import Any
from typing import Union

from httpx import Response


def check_response(
    res: Response,
    expected_status_code: Union[int, list[int]] = 200,
) -> Union[list[Any], dict[str, Any], str, int, float, bool]:
    """
    Check the validity of the http response from fractal server

    If the status code of the response is not one of the expected values, print
    the error to stderr and terminate with exit status 1.
    Some errors (422 errors due for failed validation of request body) are
    handled in a specific way, to make their error message more readable.

    Args:
        res: Response from `fractal-server`.
        expected_status_code: Expected status code(s).

    Returns:
        The output of `res.json()`.
    """

    try:
        data = res.json()
    except JSONDecodeError:
        data = {}

    # Also allow a list of expected status codes
    if isinstance(expected_status_code, list):
        expected_status_codes = expected_status_code
    else:
        expected_status_codes = [expected_status_code]

    logging.debug(res.status_code)
    if res.status_code not in expected_status_codes:

        logging.error(f"Server returned {res.status_code}")
        logging.error(
            f"Original request: {res._request.method} {res._request.url}"
        )
        logging.error(
            f"Original payload: {res._request._content.decode('utf-8')}"
        )

        error_msg = data

        # Detect whether the error is due to failed request-body validation,
        # and make the error message more readable
        if (
            res.status_code == 422
            and isinstance(data, dict)
            and list(data.keys()) == ["detail"]
            and isinstance(data["detail"], list)
            and len(data["detail"]) == 1
            and isinstance(data["detail"][0], dict)
            and set(data["detail"][0].keys()) == {"msg", "type", "loc"}
        ):
            msg = data["detail"][0]["msg"]
            _type = data["detail"][0]["type"]
            loc = data["detail"][0]["loc"]
            error_msg = f"\n\tmsg: {msg}\n\ttype: {_type}\n\tloc: {loc}"

        logging.error(f"Server error message: {error_msg}\n")
        logging.error("Terminating.\n")
        sys.exit(1)

    return data



from typing import Optional
from typing import Union

from ..authclient import AuthClient
from ..config import settings
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response


async def user_register(
    client: AuthClient,
    *,
    new_email: str,
    new_password: Optional[str] = None,
    slurm_user: Optional[str] = None,
    cache_dir: Optional[str] = None,
    username: Optional[str] = None,
    superuser: bool = False,
    batch: bool = False,
) -> Union[RichJsonInterface, PrintInterface]:

    new_user = dict(
        email=new_email,
        password=new_password,
    )
    if slurm_user:
        new_user["slurm_user"] = slurm_user
    if cache_dir:
        new_user["cache_dir"] = cache_dir
    if username:
        new_user["username"] = username

    from getpass import getpass

    if new_password is None:
        new_password = getpass()
        confirm_new_password = getpass("Confirm password: ")
        if new_password == confirm_new_password:
            new_user["password"] = new_password
        else:
            return PrintInterface(retcode=1, data="Passwords do not match.")

    res = await client.post(
        f"{settings.FRACTAL_SERVER}/auth/register/", json=new_user
    )
    data = check_response(res, expected_status_code=201)

    if superuser:
        user_id = data["id"]
        res = await client.patch(
            f"{settings.FRACTAL_SERVER}/auth/users/{user_id}/",
            json={"is_superuser": True},
        )
        data = check_response(res, expected_status_code=200)

    if batch:
        return PrintInterface(retcode=0, data=data["id"])
    else:
        return RichJsonInterface(retcode=0, data=data)


async def user_list(client: AuthClient) -> RichJsonInterface:
    res = await client.get(f"{settings.FRACTAL_SERVER}/auth/users/")
    users = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=users)


async def user_show(client: AuthClient, *, user_id: str) -> RichJsonInterface:
    res = await client.get(f"{settings.FRACTAL_SERVER}/auth/users/{user_id}/")
    user = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=user)


async def user_edit(
    client: AuthClient,
    *,
    user_id: str,
    new_email: Optional[str] = None,
    new_password: Optional[str] = None,
    new_slurm_user: Optional[str] = None,
    new_cache_dir: Optional[str] = None,
    new_username: Optional[str] = None,
    make_superuser: bool = False,
    remove_superuser: bool = False,
) -> Union[RichJsonInterface, PrintInterface]:

    user_update = dict()
    if new_email is not None:
        user_update["email"] = new_email
    if new_password is not None:
        user_update["password"] = new_password
    if make_superuser:
        user_update["is_superuser"] = True
    if remove_superuser:
        user_update["is_superuser"] = False
    if new_cache_dir is not None:
        user_update["cache_dir"] = new_cache_dir
    if new_slurm_user is not None:
        user_update["slurm_user"] = new_slurm_user
    if new_username is not None:
        user_update["username"] = new_username

    if not user_update:
        return PrintInterface(retcode=1, data="Nothing to update")
    res = await client.patch(
        f"{settings.FRACTAL_SERVER}/auth/users/{user_id}/", json=user_update
    )
    new_user = check_response(res, expected_status_code=200)

    return RichJsonInterface(retcode=0, data=new_user)


async def user_whoami(
    client: AuthClient, *, batch: bool
) -> Union[RichJsonInterface, PrintInterface]:
    res = await client.get(f"{settings.FRACTAL_SERVER}/auth/current-user/")
    user = check_response(res, expected_status_code=200)

    if batch:
        return PrintInterface(retcode=0, data=user["id"])
    else:
        return RichJsonInterface(retcode=0, data=user)



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
Tommaso Comparin <tommaso.comparin@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import logging
from os import getenv
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv

load_dotenv(".fractal.env")


class Settings:
    def __init__(self):

        self.FRACTAL_LOGGING_LEVEL: int = getenv(
            "FRACTAL_LOGGING_LEVEL", logging.INFO
        )

        self.FRACTAL_USER: Optional[str] = getenv("FRACTAL_USER")
        self.FRACTAL_PASSWORD: Optional[str] = getenv("FRACTAL_PASSWORD")

        self.FRACTAL_SERVER: str = getenv(
            "FRACTAL_SERVER", "http://localhost:8000"
        )
        self.FRACTAL_CACHE_PATH: str = getenv(
            "FRACTAL_CACHE_PATH", str(Path.home() / ".cache/fractal")
        )

    @property
    def BASE_URL(self) -> str:
        return f"{self.FRACTAL_SERVER}/api/v1"


settings = Settings()



import json
from pathlib import Path
from typing import Any
from typing import Optional

import packaging.version

from ..authclient import AuthClient
from ..config import settings
from ..response import check_response

TASKS_CACHE_FILENAME = "tasks"


def _loose_version_parse(v: str) -> packaging.version.Version:
    """
    Catch `InvalidVersion` error and return `Version("0")`.

    This function is used in the comparison of different version strings. If a
    version cannot be parsed correctly, then it should not be considered the
    "latest"; we obtain this behavior by returning the "0" version when
    version-string parsing fails.

    Args:
        v: Version string (e.g. `0.10.0a2`).

    Returns:
        A `Version` object, parsed with `packaging.version.parse`
    """
    try:
        return packaging.version.parse(v)
    except packaging.version.InvalidVersion:
        return packaging.version.parse("0")


class FractalCacheError(RuntimeError):
    """
    Custom error raised by functions of this module
    """

    pass


# Define a useful type
_TaskList = list[dict[str, Any]]


async def _fetch_task_list(client: AuthClient) -> _TaskList:
    """
    Fetch task list through an API request.
    """
    res = await client.get(f"{settings.BASE_URL}/task/")
    task_list = check_response(res, expected_status_code=200)
    return task_list


def _sort_task_list(task_list: _TaskList) -> _TaskList:
    """
    Sort tasks according to their (owner, name, version) attributes.
    """
    new_task_list = sorted(
        task_list,
        key=lambda task: (
            task["owner"] or "",
            task["name"],
            task["version"] or "",
        ),
    )
    return new_task_list


def _write_task_list_to_cache(task_list: _TaskList) -> None:
    """
    Write task list to cache file
    """
    cache_dir = Path(f"{settings.FRACTAL_CACHE_PATH}")
    cache_dir.mkdir(parents=True, exist_ok=True)
    with (cache_dir / TASKS_CACHE_FILENAME).open("w") as f:
        json.dump(task_list, f, indent=4)


async def refresh_task_cache(client: AuthClient) -> _TaskList:
    """
    Return task list after fetching it, sorting it and writing to cache file.
    """
    task_list = await _fetch_task_list(client)
    task_list = _sort_task_list(task_list)
    _write_task_list_to_cache(task_list)
    return task_list


def _get_matching_tasks(
    task_list: _TaskList,
    *,
    name: str,
    version: Optional[str] = None,
) -> _TaskList:
    """
    Given a task list, extract all the tasks matching some conditions.
    """

    def _condition(_task):
        if _task["name"] == name:
            if (version is None) or (_task["version"] == version):
                return True
            return False
        else:
            return False

    return [_task for _task in task_list if _condition(_task)]


def _format_task_list(task_list: _TaskList) -> str:
    """
    Helper function to print a formatted task list with only a few task
    attributes, to be used in error messages.
    """
    header = "  ID, Name, Version, Owner, Source"
    formatted_list = "\n".join(
        [
            (
                f'  {task["id"]}, "{task["name"]}", {task["version"]}, '
                f'{task.get("owner")}, {task["source"]}'
            )
            for task in task_list
        ]
    )
    return f"{header}\n{formatted_list}"


def _search_in_task_list(
    *,
    task_list: _TaskList,
    name: str,
    version: Optional[str] = None,
) -> int:
    """
    Search for a single task in `task_list` based on the provided `name`
    and `version`, and return its `id`.

    If `version` is not provided, use the maximum available version (that is,
    the latest version).

    If the task is not found or is not unique, raise a `FractalCacheError`.
    """
    matching_task_list = _get_matching_tasks(
        task_list, name=name, version=version
    )
    formatted_matching_task_list = _format_task_list(matching_task_list)

    if len(matching_task_list) == 0:
        formatted_task_list = _format_task_list(task_list)
        if version is not None:
            raise FractalCacheError(
                f'There is no task with (name, version)=("{name}", {version}) '
                f"in the following task list:\n{formatted_task_list}\n"
            )
        else:
            raise FractalCacheError(
                f'There is no task with name "{name}" '
                f"in the following task list:\n{formatted_task_list}\n"
            )
    elif len(matching_task_list) == 1:
        return matching_task_list[0]["id"]
    else:  # i.e. len(matching_task_list) > 1
        if version is not None:
            raise FractalCacheError(
                f"Multiple tasks with version {version} in the following "
                f"task list:\n{formatted_matching_task_list}"
                "Please make your request more specific.\n"
            )
        else:  # i.e. version is None
            if any(task["version"] is None for task in matching_task_list):
                raise FractalCacheError(
                    "Cannot determine the latest version in the following "
                    f"task list:\n{formatted_matching_task_list}"
                    "Please make your request more specific.\n"
                )
            available_versions = [
                _task["version"] for _task in matching_task_list
            ]
            max_version = max(available_versions, key=_loose_version_parse)
            max_version_tasks = [
                _task
                for _task in matching_task_list
                if _task["version"] == max_version
            ]
            formatted_matching_task_list = _format_task_list(max_version_tasks)
            if len(max_version_tasks) == 1:
                return max_version_tasks[0]["id"]
            else:
                raise FractalCacheError(
                    "Multiple tasks with latest version "
                    f"({max_version}) in the following task "
                    f"list:\n{formatted_matching_task_list}"
                    "Please make your request more specific.\n"
                )


async def get_task_id_from_cache(
    client: AuthClient, task_name: str, version: Optional[str] = None
) -> int:
    """
    Retrieve the `id` of a task from the cache based on the provided
    `task_name` and `version`.

    If `version` is not provided, the latest (i.e. maximum) available version
    is used.

    Return the `id` of the single matching task, if found.

    If the task is not found or is not unique, re-try after refreshing the
    cache, and then raise a `FractalCacheError`.
    """

    # If cache is missing, create it
    cache_dir = Path(f"{settings.FRACTAL_CACHE_PATH}")
    cache_file = cache_dir / TASKS_CACHE_FILENAME
    if cache_file.exists():
        with cache_file.open("r") as f:
            task_list = json.load(f)
        already_refreshed_cache = False
    else:
        task_list = await refresh_task_cache(client)
        already_refreshed_cache = True

    try:
        task_id = _search_in_task_list(
            task_list=task_list,
            name=task_name,
            version=version,
        )
    except FractalCacheError as e:
        if already_refreshed_cache:
            # Cache is already up to date, fail
            raise e
        else:
            # Cache may be out-of-date, refresh it and try again
            task_list = await refresh_task_cache(client)
            task_id = _search_in_task_list(
                task_list=task_list,
                name=task_name,
                version=version,
            )
    return task_id



from httpx import AsyncClient

from ..authclient import AuthClient
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ._dataset import delete_dataset
from ._dataset import delete_resource
from ._dataset import get_dataset
from ._dataset import get_dataset_history
from ._dataset import get_dataset_status
from ._dataset import patch_dataset
from ._dataset import post_dataset
from ._dataset import post_resource
from ._job import get_job
from ._job import get_job_list
from ._job import get_job_logs
from ._job import stop_job
from ._project import delete_project
from ._project import get_project
from ._project import get_project_list
from ._project import patch_project
from ._project import post_project
from ._task import delete_task
from ._task import get_task_list
from ._task import patch_task
from ._task import post_task
from ._task import task_collect_pip
from ._task import task_collection_check
from ._user import user_edit
from ._user import user_list
from ._user import user_register
from ._user import user_show
from ._user import user_whoami
from ._workflow import delete_workflow
from ._workflow import delete_workflowtask
from ._workflow import get_workflow
from ._workflow import get_workflow_list
from ._workflow import patch_workflow
from ._workflow import patch_workflowtask
from ._workflow import post_workflow
from ._workflow import post_workflowtask
from ._workflow import workflow_apply
from ._workflow import workflow_export
from ._workflow import workflow_import
from fractal_client import __VERSION__


class NoCommandError(ValueError):
    pass


def get_kwargs(_parameters, _kwargs):
    return {k: _kwargs.get(k) for k in _parameters if k in _kwargs}


async def project(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:

    if subcmd == "new":
        parameters = ["name"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await post_project(client, batch=batch, **function_kwargs)
    elif subcmd == "show":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_project(client, **function_kwargs)
    elif subcmd == "list":
        iface = await get_project_list(client)
    elif subcmd == "edit":
        parameters = [
            "project_id",
            "new_name",
            "make_read_only",
            "remove_read_only",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await patch_project(client, **function_kwargs)
    elif subcmd == "add-dataset":
        parameters = [
            "project_id",
            "dataset_name",
            "metadata",
            "type",
            "make_read_only",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await post_dataset(client, batch=batch, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await delete_project(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command project {subcmd} not found")

    return iface


async def dataset(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:
    if subcmd == "show":
        parameters = ["project_id", "dataset_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_dataset(client, **function_kwargs)
    elif subcmd == "add-resource":
        parameters = ["project_id", "dataset_id", "path"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await post_resource(client, batch=batch, **function_kwargs)
    elif subcmd == "rm-resource":
        parameters = ["project_id", "dataset_id", "resource_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await delete_resource(client, **function_kwargs)
    elif subcmd == "edit":
        parameters = [
            "project_id",
            "dataset_id",
            "new_name",
            "new_type",
            "meta_file",
            "make_read_only",
            "remove_read_only",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await patch_dataset(client, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["project_id", "dataset_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await delete_dataset(client, **function_kwargs)
    elif subcmd == "history":
        parameters = ["project_id", "dataset_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_dataset_history(client, **function_kwargs)
    elif subcmd == "status":
        parameters = ["project_id", "dataset_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_dataset_status(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command dataset {subcmd} not found")
    return iface


async def task(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:

    if subcmd == "list":
        iface = await get_task_list(client)
    elif subcmd == "collect":
        parameters = [
            "package",
            "package_version",
            "python_version",
            "package_extras",
            "pinned_dependency",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await task_collect_pip(client, batch=batch, **function_kwargs)
    elif subcmd == "check-collection":
        parameters = ["state_id", "include_logs", "do_not_separate_logs"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await task_collection_check(client, **function_kwargs)
    elif subcmd == "new":
        parameters = [
            "name",
            "command",
            "source",
            "input_type",
            "output_type",
            "version",
            "meta_file",
            "args_schema",
            "args_schema_version",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await post_task(client, batch=batch, **function_kwargs)
    elif subcmd == "edit":
        parameters = [
            "id",
            "name",
            "version",
            "new_name",
            "new_command",
            "new_input_type",
            "new_output_type",
            "new_version",
            "meta_file",
            "new_args_schema",
            "new_args_schema_version",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await patch_task(client, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["id", "name", "version"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await delete_task(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command task {subcmd} not found")
    return iface


async def workflow(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:
    if subcmd == "show":
        parameters = ["project_id", "workflow_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_workflow(client, **function_kwargs)
    elif subcmd == "new":
        parameters = ["name", "project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await post_workflow(client, batch=batch, **function_kwargs)
    elif subcmd == "list":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_workflow_list(client, batch=batch, **function_kwargs)
    elif subcmd == "edit":
        parameters = ["project_id", "workflow_id", "new_name"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await patch_workflow(client, **function_kwargs)
    elif subcmd == "delete":
        parameters = ["project_id", "workflow_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await delete_workflow(client, **function_kwargs)
    elif subcmd == "add-task":
        parameters = [
            "project_id",
            "workflow_id",
            "task_id",
            "task_name",
            "task_version",
            "order",
            "args_file",
            "meta_file",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await post_workflowtask(client, batch=batch, **function_kwargs)
    elif subcmd == "edit-task":
        parameters = [
            "project_id",
            "workflow_id",
            "workflow_task_id",
            "args_file",
            "meta_file",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await patch_workflowtask(client, **function_kwargs)
    elif subcmd == "rm-task":
        parameters = ["project_id", "workflow_id", "workflow_task_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await delete_workflowtask(client, **function_kwargs)
    elif subcmd == "apply":
        parameters = [
            "project_id",
            "workflow_id",
            "input_dataset_id",
            "output_dataset_id",
            "worker_init",
            "first_task_index",
            "last_task_index",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await workflow_apply(client, batch=batch, **function_kwargs)
    elif subcmd == "import":
        parameters = ["project_id", "json_file"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await workflow_import(client, batch=batch, **function_kwargs)
    elif subcmd == "export":
        parameters = ["project_id", "workflow_id", "json_file"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await workflow_export(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command workflow {subcmd} not found")
    return iface


async def job(
    client: AuthClient,
    subcmd: str,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:
    if subcmd == "list":
        parameters = ["project_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_job_list(client, batch=batch, **function_kwargs)
    elif subcmd == "show":
        parameters = ["project_id", "job_id", "do_not_separate_logs"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_job(client, batch=batch, **function_kwargs)
    elif subcmd == "download-logs":
        parameters = ["project_id", "job_id", "output_folder"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await get_job_logs(client, **function_kwargs)
    elif subcmd == "stop":
        parameters = ["project_id", "job_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await stop_job(client, **function_kwargs)
    else:
        raise NoCommandError(f"Command job {subcmd} not found")
    return iface


async def version(client: AsyncClient, **kwargs) -> PrintInterface:
    res = await client.get(f"{settings.FRACTAL_SERVER}/api/alive/")
    data = res.json()

    return PrintInterface(
        retcode=0,
        data=(
            f"Fractal client\n\tversion: {__VERSION__}\n"
            "Fractal server:\n"
            f"\turl: {settings.FRACTAL_SERVER}"
            f"\tversion: {data['version']}"
        ),
    )


async def user(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "register":
        parameters = [
            "new_email",
            "new_password",
            "cache_dir",
            "slurm_user",
            "username",
            "superuser",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await user_register(client, batch=batch, **function_kwargs)
    elif subcmd == "list":
        iface = await user_list(client)
    elif subcmd == "show":
        parameters = ["user_id"]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await user_show(client, **function_kwargs)
    elif subcmd == "edit":
        parameters = [
            "user_id",
            "new_email",
            "new_password",
            "new_cache_dir",
            "new_slurm_user",
            "new_username",
            "make_superuser",
            "remove_superuser",
        ]
        function_kwargs = get_kwargs(parameters, kwargs)
        iface = await user_edit(client, **function_kwargs)
    elif subcmd == "whoami":
        iface = await user_whoami(client, batch=batch)
    else:
        raise NoCommandError(f"Command user {subcmd} not found")

    return iface



import json
import logging
import sys
from pathlib import Path
from typing import Optional

from ..authclient import AuthClient
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response
from ._aux_task_caching import FractalCacheError
from ._aux_task_caching import get_task_id_from_cache


async def post_workflow(
    client: AuthClient, *, name: str, project_id: int, batch: bool = False
) -> BaseInterface:
    workflow = dict(
        name=name,
        project_id=project_id,
    )
    res = await client.post(
        f"{settings.BASE_URL}/project/{project_id}/workflow/",
        json=workflow,
    )
    workflow = check_response(res, expected_status_code=201)
    if batch:
        return PrintInterface(retcode=0, data=workflow["id"])
    else:
        return RichJsonInterface(retcode=0, data=workflow)


async def get_workflow_list(
    client: AuthClient, *, project_id: int, batch: bool = False
) -> RichJsonInterface:

    res = await client.get(
        f"{settings.BASE_URL}/project/{project_id}/workflow/"
    )
    workflow_list = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=workflow_list)


async def delete_workflow(
    client: AuthClient, *, project_id: int, workflow_id: int
) -> BaseInterface:
    res = await client.delete(
        f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def get_workflow(
    client: AuthClient, *, project_id: int, workflow_id: int
) -> RichJsonInterface:
    res = await client.get(
        f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/"
    )
    workflow = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=workflow)


async def post_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    task_id: Optional[int] = None,
    task_name: Optional[str] = None,
    task_version: Optional[str] = None,
    batch: bool = False,
    order: Optional[int] = None,
    args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
) -> RichJsonInterface:

    if task_id:
        if task_version:
            logging.error(
                "Too many arguments: cannot provide both "
                "`task_id` and `task_version`."
            )
            sys.exit(1)
    else:
        try:
            task_id = await get_task_id_from_cache(
                client=client, task_name=task_name, version=task_version
            )
        except FractalCacheError as e:
            print(e)
            sys.exit(1)

    if order is None:
        workflow_task = dict()
    else:
        workflow_task = dict(order=order)
    if args_file:
        with Path(args_file).open("r") as f:
            args = json.load(f)
            workflow_task["args"] = args
    if meta_file:
        with Path(meta_file).open("r") as f:
            meta = json.load(f)
            workflow_task["meta"] = meta

    res = await client.post(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"workflow/{workflow_id}/wftask/"
            f"?{task_id=}"
        ),
        json=workflow_task,
    )
    workflow_task = check_response(res, expected_status_code=201)

    if batch:
        return PrintInterface(retcode=0, data=str(workflow_task["id"]))
    else:
        return RichJsonInterface(retcode=0, data=workflow_task)


async def patch_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    workflow_task_id: int,
    args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
) -> RichJsonInterface:

    # Check that at least one of args_file or meta_file was given (note: it
    # would be reasonable to check it in the parser, but we are not aware of a
    # method within argparse).
    if not (args_file or meta_file):
        raise ValueError(
            "At least one of {args_file,meta_file} arguments is required"
        )

    # Combine args/meta payload
    payload = {}
    if args_file:
        with Path(args_file).open("r") as f:
            args = json.load(f)
            payload["args"] = args
    if meta_file:
        with Path(meta_file).open("r") as f:
            meta = json.load(f)
            payload["meta"] = meta

    res = await client.patch(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"workflow/{workflow_id}/wftask/{workflow_task_id}/"
        ),
        json=payload,
    )
    workflow_task = check_response(res, expected_status_code=200)

    return RichJsonInterface(retcode=0, data=workflow_task)


async def delete_workflowtask(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    workflow_task_id: int,
) -> BaseInterface:

    res = await client.delete(
        f"{settings.BASE_URL}/project/{project_id}/"
        f"workflow/{workflow_id}/wftask/{workflow_task_id}/"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def patch_workflow(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    new_name: str,
) -> BaseInterface:

    workflow_update = dict(name=new_name)

    res = await client.patch(
        f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/",
        json=workflow_update,
    )
    new_workflow = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=new_workflow)


async def workflow_apply(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    input_dataset_id: int,
    output_dataset_id: int,
    first_task_index: Optional[int] = None,
    last_task_index: Optional[int] = None,
    worker_init: Optional[str] = None,
    batch: bool = False,
) -> BaseInterface:
    apply_wf_create = dict(
        workflow_id=workflow_id,
        input_dataset_id=input_dataset_id,
        output_dataset_id=output_dataset_id,
    )
    # Prepare ApplyWorkflowCreate object, without None attributes
    if worker_init is not None:
        apply_wf_create["worker_init"] = worker_init
    if first_task_index is not None:
        apply_wf_create["first_task_index"] = first_task_index
    if last_task_index is not None:
        apply_wf_create["last_task_index"] = last_task_index

    # Prepare query parameters
    query_parameters = (
        f"input_dataset_id={input_dataset_id}"
        f"&output_dataset_id={output_dataset_id}"
    )

    res = await client.post(
        (
            f"{settings.BASE_URL}/project/{project_id}/workflow/{workflow_id}/"
            f"apply/?{query_parameters}"
        ),
        json=apply_wf_create,
    )
    apply_wf_read = check_response(res, expected_status_code=202)

    if batch:
        return PrintInterface(retcode=0, data=apply_wf_read["id"])
    else:
        return RichJsonInterface(retcode=0, data=apply_wf_read)


async def workflow_import(
    client: AuthClient, *, project_id: int, json_file: str, batch: bool = False
) -> BaseInterface:
    with Path(json_file).open("r") as f:
        workflow = json.load(f)

    res = await client.post(
        f"{settings.BASE_URL}/project/{project_id}/workflow/import/",
        json=workflow,
    )
    wf_read = check_response(res, expected_status_code=201)

    warnings = [
        workflow_task["task"]["source"]
        for workflow_task in wf_read["task_list"]
        if workflow_task["task"]["owner"]
    ]
    if warnings:
        sources_str = ", ".join([f"'{s}'" for s in warnings])
        logging.warning(
            "This workflow includes custom tasks (the ones with sources: "
            f"{sources_str}), which are not meant to be portable; "
            "importing this workflow may not work as expected."
        )

    if batch:
        datastr = f"{wf_read['id']}"
        for wftask in wf_read["task_list"]:
            datastr += f" {wftask['id']}"
        return PrintInterface(retcode=0, data=datastr)
    else:
        return RichJsonInterface(retcode=0, data=wf_read)


async def workflow_export(
    client: AuthClient,
    *,
    project_id: int,
    workflow_id: int,
    json_file: str,
) -> BaseInterface:
    res = await client.get(
        (
            f"{settings.BASE_URL}/project/{project_id}/"
            f"workflow/{workflow_id}/export/"
        ),
    )
    workflow = check_response(res, expected_status_code=200)

    warnings = [
        workflow_task["task"]["source"]
        for workflow_task in workflow["task_list"]
        if not workflow_task["task"]["source"].startswith(
            ("pip_local:", "pip_remote:")
        )
    ]
    if warnings:
        sources_str = ", ".join([f"'{s}'" for s in warnings])
        logging.warning(
            "This workflow includes custom tasks (the ones with sources: "
            f"{sources_str}), which are not meant to be portable; "
            "re-importing this workflow may not work as expected."
        )

    with Path(json_file).open("w") as f:
        json.dump(workflow, f, indent=2)
    return PrintInterface(
        retcode=0, data=f"Workflow {workflow_id} exported at {json_file}"
    )
